#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "ExprEvaluator.h"
#include "ExpressionLexer.h"
#include "stack_ve.h"
#include "stack_ve_sym.h"

//Stack
boolean getOperands(OperandStack evalStack, OperandType *operand1, OperandType *operand2);

//Valutazione
OperandType evaluate(OperandType operand1, OperandType operand2, SymbolType op);

//Gestione operatori
boolean isOperator(SymbolType symbol);
int getOperatorPriority(SymbolType op);
int compareOpPriority(SymbolType op1, SymbolType op2);

void appendToken(char expr[], Token t);

boolean postfixEval(char expr[], OperandType *value)
{
    boolean result = false, noError = true;
    OperandStack evalStack = newOperandStack();
    ExpressionLexer *lexer = newExpressionLexer(expr);
    while (readNextTokenFromLex(lexer) && noError)
    {
        SymbolType type = getCurrentTokenType(lexer);
        switch (type)
        {
        case Operand:
            {
                Token t;
                OperandType operand;
                getCurrentTokenFromLex(lexer, t);
                operand = convertToOperand(t);
                pushOperand(evalStack, operand);
                break;
            }
        case OperatorPlus:
        case OperatorMinus:
        case OperatorMul:
        case OperatorDiv:
            {
                OperandType operand1, operand2;
                noError = getOperands(evalStack, &operand1, &operand2);
                if (noError)
                {
                    pushOperand(evalStack, evaluate(operand1, operand2, type));
                }
                break;
            }
        default:
            assert(false);
            break;
        }
    }
    if (noError && !isEmptyOperandStack(evalStack))
    {
        *value = popOperand(evalStack);
        result = isEmptyOperandStack(evalStack);
    }
    destroyOperandStack(evalStack);
    destroyExpressionLexer(lexer);
    return result;
}

boolean getOperands(OperandStack evalStack, OperandType *operand1, OperandType *operand2)
{
    if (isEmptyOperandStack(evalStack))
        return false;
    *operand2 = popOperand(evalStack);
    if (isEmptyOperandStack(evalStack))
        return false;
    *operand1 = popOperand(evalStack);
    return true;
}

void appendToken(char expr[], Token t)
{
    int len = strlen(expr);
    if (len > 0)
    {
        strcat(expr, " ");
    }
    strcat(expr, t);
}

void appendOperator(char expr[], SymbolType symbol)
{
    int len = strlen(expr);
    if (len != 0)
    {
        expr[len++] = ' ';
    }
    switch (symbol)
    {
    case OperatorPlus:
        expr[len] = '+';
        break;
    case OperatorMinus:
        expr[len] = '-';
        break;
    case OperatorMul:
        expr[len] = '*';
        break;
    case OperatorDiv:
        expr[len] = '/';
        break;
    default:
        assert(false);  //Errore!
        break;
    }
    expr[++len] = '\0';
}

boolean infixToPostfix(char inExpr[], char outExpr[])
{
    boolean result, noError = true;
    SymbolStack symStack = newSymbolStack();
    ExpressionLexer *lexer = newExpressionLexer(inExpr);
    Token t;
    outExpr[0] = '\0';

    while (noError && readNextTokenFromLex(lexer))
    {
        switch (getCurrentTokenType(lexer))
        {
        case Operand:
            {
                getCurrentTokenFromLex(lexer, t);
                appendToken(outExpr, t);
                break;
            }
        case OperatorPlus:
        case OperatorMinus:
        case OperatorMul:
        case OperatorDiv:
            {
                SymbolType symbol = getCurrentTokenType(lexer);
                while (!isEmptySymbolStack(symStack) && isOperator(peekSymbol(symStack)) && compareOpPriority(symbol, peekSymbol(symStack)) <= 0)
                {
                    SymbolType firstSymbolInStack = popSymbol(symStack);    //Estrae l'operatore gi letto con peek
                    appendOperator(outExpr, firstSymbolInStack);
                }
                pushSymbol(symbol, symStack);
                break;
            }
        case ParOpen:
            {
                SymbolType symbol = getCurrentTokenType(lexer);
                pushSymbol(symbol, symStack);
                break;
            }
        case ParClose:
            {
                SymbolType firstSymbolInStack;
                if (!isEmptySymbolStack(symStack))
                {
                    firstSymbolInStack = popSymbol(symStack);
                    while (noError && firstSymbolInStack != ParOpen)
                    {
                        appendOperator(outExpr, firstSymbolInStack);
                        noError = !isEmptySymbolStack(symStack);
                        if (noError)
                        {
                            firstSymbolInStack = popSymbol(symStack);
                        }
                    }
                }
                break;
            }
        }
    }
    while (noError && !isEmptySymbolStack(symStack))
    {
        SymbolType op = popSymbol(symStack);
        noError = isOperator(op);
        appendOperator(outExpr, op);
    }
    result = noError && isEmptySymbolStack(symStack);
    destroySymbolStack(symStack);
    destroyExpressionLexer(lexer);
    return result;
}

boolean infixEval(char expr[], int *value)
{
    boolean result;
    OperandStack evalStack = newOperandStack();
    SymbolStack symStack = newSymbolStack();
    ExpressionLexer *lexer = newExpressionLexer(expr);

    while (readNextTokenFromLex(lexer))
    {
        switch (getCurrentTokenType(lexer))
        {
        case Operand:
            {
                Token t;
                OperandType operand;
                getCurrentTokenFromLex(lexer, t);
                operand = convertToOperand(t);
                pushOperand(evalStack, operand);
                break;
            }
        case OperatorPlus:
        case OperatorMinus:
        case OperatorMul:
        case OperatorDiv:
            {
                SymbolType symbol = getCurrentTokenType(lexer);
                while (!isEmptySymbolStack(symStack) && isOperator(peekSymbol(symStack)) && compareOpPriority(symbol, peekSymbol(symStack)) <= 0)
                {
                    OperandType operand1, operand2;
                    SymbolType firstSymbolInStack = popSymbol(symStack);    //Estrae l'operatore gi letto con peek
                    if (!getOperands(evalStack, &operand1, &operand2))
                    {
                        destroyOperandStack(evalStack);
                        destroySymbolStack(symStack);
                        destroyExpressionLexer(lexer);
                        return false;
                    }
                    pushOperand(evalStack, evaluate(operand1, operand2, firstSymbolInStack));
                }
                pushSymbol(symbol, symStack);
                break;
            }
         case ParOpen:
             {
                 SymbolType symbol = getCurrentTokenType(lexer);
                 pushSymbol(symbol, symStack);
                 break;
             }
         case ParClose:
             {
                 SymbolType firstSymbolInStack;
                 for (firstSymbolInStack = popSymbol(symStack); firstSymbolInStack != ParOpen; firstSymbolInStack = popSymbol(symStack))
                 {
                     OperandType operand1, operand2;
                     if (!getOperands(evalStack, &operand1, &operand2))
                     {
                         destroyOperandStack(evalStack);
                         destroySymbolStack(symStack);
                         destroyExpressionLexer(lexer);
                         return false;
                     }
                     pushOperand(evalStack, evaluate(operand1, operand2, firstSymbolInStack));
                 }
                 break;
             }
        }
    }
    while (!isEmptySymbolStack(symStack))
    {
        SymbolType op = popSymbol(symStack);
        OperandType operand1, operand2;
        if (!getOperands(evalStack, &operand1, &operand2))
        {
            destroyOperandStack(evalStack);
            destroySymbolStack(symStack);
            destroyExpressionLexer(lexer);
            return false;
        }
        pushOperand(evalStack, evaluate(operand1, operand2, op));
    }
    if (!isEmptyOperandStack(evalStack))
        *value = popOperand(evalStack);
    result = isEmptySymbolStack(symStack) && isEmptyOperandStack(evalStack);
    destroyOperandStack(evalStack);
    destroySymbolStack(symStack);
    destroyExpressionLexer(lexer);
    return result;
}

OperandType evaluate(OperandType operand1, OperandType operand2, SymbolType op)
{
    switch (op)
    {
    case OperatorPlus:
        return doAdd(operand1, operand2);
    case OperatorMinus:
        return  doSub(operand1, operand2);
    case OperatorMul:
        return doMul(operand1, operand2);
    case OperatorDiv:
        return doDiv(operand1, operand2);
    default:
        assert(false);  //Operatore non valido
        return 0;
    }
}

boolean isOperator(SymbolType symbol)
{
    return symbol == OperatorPlus || symbol == OperatorMinus || symbol == OperatorMul || symbol == OperatorDiv;
}

int getOperatorPriority(SymbolType op)
{
    switch (op)
    {
    case OperatorPlus:
    case OperatorMinus:
        return 1;
    case OperatorMul:
    case OperatorDiv:
        return 2;
    default:
        assert(false);  //Operatore non valido
        return 0;
    }
}

int compareOpPriority(SymbolType op1, SymbolType op2)
{
    return getOperatorPriority(op1) - getOperatorPriority(op2);
}

void postfixEvalTest()
{
    int value;
    assert(postfixEval("10 2 + 1 4 * /", &value));
    assert(value == 3);
    assert(!postfixEval("10 2 4 +", &value));
}

void infixToPostfixTest()
{
    char outExpr[30];
    assert(infixToPostfix("10 - 2 * 3", outExpr));
    assert(strcmp("10 2 3 * -", outExpr) == 0);

    assert(infixToPostfix("10 * 2 - 3", outExpr));
    assert(strcmp("10 2 * 3 -", outExpr) == 0);

    assert(infixToPostfix("2 * ( 3 + 4 )", outExpr));
    assert(strcmp("2 3 4 + *", outExpr) == 0);

    assert(infixToPostfix("10 - 2 - 3 * 2 / ( 2 * 1 )", outExpr));
    assert(strcmp("10 2 - 3 2 * 2 1 * / -", outExpr) == 0);
}

void infixEvalTest()
{
    int value;

    assert(infixEval("10 - 2 * 3", &value));
    assert(value == 4);

    assert(infixEval("10 * 2 - 3", &value));
    assert(value == 17);

    assert(infixEval("2 * ( 3 + 4 )", &value));
    assert(value == 14);

    assert(infixEval("10 - 2 - 3 * 2 / ( 2 * 1 )", &value));
    assert(value == 5);
}